import re
from enum import Enum, unique, auto
from typing import Union

URL_PATTERN = r'[A-Za-z0-9]+://[A-Za-z0-9%-_]+(/[A-Za-z0-9%-_])*(#|\\?)[A-Za-z0-9%-_&=]*'


def remove_urls(text: str) -> str:
    """ Remove any url in the text.

    :param text: The text to remove urls.
    :return: The same text but without urls.
    """
    return re.sub(URL_PATTERN, '', text)


def clean_text(text: str, lower: bool = True, url: bool = True) -> str:
    """ Remove punctuation symbols, urls and convert to lower.

    :param text: The text to clean.
    :param lower: If you want to convert to lower.
    :param url: If you want to remove urls.
    :return: The cleaned text.
    """
    text = text.lower() if lower else text
    text = remove_urls(text) if url else text
    text = re.sub(r'[^A-Za-z0-9 -]', ' ', text)
    return re.sub(r'[\s]+', ' ', text).strip()


def color(r: int, g: int, b: int) -> str:
    """ Get a ANSI code for a given (R, G, B) foreground color.

    :param r: A value between 0 and 255 for the red component color.
    :param g: A value between 0 and 255 for the green component color.
    :param b: A value between 0 and 255 for the blue component color.
    :return: A string that represents the ANSI code color.
    """
    return f'\033[38{_color(r, g, b)}m'


def bg_color(r: int, g: int, b: int) -> str:
    """ Get a ANSI code for a given (R, G, B) background color.

    :param r: A value between 0 and 255 for the red component color.
    :param g: A value between 0 and 255 for the green component color.
    :param b: A value between 0 and 255 for the blue component color.
    :return: A string that represents the ANSI code background color.
    """
    return f'\033[48{_color(r, g, b)}m'


def un_color(r: int, g: int, b: int) -> str:
    """ Get a ANSI code for a given (R, G, B) underline color.

    :param r: A value between 0 and 255 for the red component color.
    :param g: A value between 0 and 255 for the green component color.
    :param b: A value between 0 and 255 for the blue component color.
    :return: A string that represents the ANSI code underline color.
    """
    return f'\033[58{_color(r, g, b)}m'


def _color(r: int, g: int, b: int) -> str:
    """ Check if the values are correct and return the part of the ANSI code that represents the (R, G, B) color.

    :param r: A value between 0 and 255 for the red component color.
    :param g: A value between 0 and 255 for the green component color.
    :param b: A value between 0 and 255 for the blue component color.
    :return: The part of the ANSI code that represents the (R, G, B) color.
    :raises ValueError: If any (R, G, B) values are not between 0 and 255.
    """
    if 0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255:
        return f';2;{r};{g};{b}'
    raise ValueError(f'The (r, g, b) values must be between 0 and 255. Current values (r={r}, g={g}, b={b}).')


@unique
class AnsiCodes(Enum):
    """ Enumeration with the most relevant ANSI codes for terminals. """
    NORMAL = '\033[0m'
    BOLD = '\033[1m'
    DIM = '\033[2m'
    ITALIC = '\033[3m'
    UNDERLINE = '\033[4m'
    SLOW_BLINK = '\033[5m'
    RAPID_BLINK = '\033[6m'
    INVERT = '\033[7m'
    HIDE = '\033[8m'
    STRIKE = '\033[9m'
    # PRIMARY = '\033[10m'
    # ALTERNATIVE = '\033[13m'
    GOTHIC = '\033[20m'
    DOUBLE = '\033[21m'
    NOT_INTENSITY = '\033[22m'
    NOT_ITALIC = '\033[23m'
    NOT_UNDERLINE = '\033[24m'
    NOT_BLINKING = '\033[25m'
    NOT_REVERSE = '\033[27m'
    NOT_HIDE = '\033[28m'
    NOT_STRIKE = '\033[29m'
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    DEFAULT = '\033[39m'
    BG_BLACK = '\033[40m'
    BG_RED = '\033[41m'
    BG_GREEN = '\033[42m'
    BG_YELLOW = '\033[43m'
    BG_BLUE = '\033[44m'
    BG_MAGENTA = '\033[45m'
    BG_CYAN = '\033[46m'
    BG_WHITE = '\033[47m'
    BG_DEFAULT = '\033[49m'
    DEF_UNDERLINE = '\033[59m'
    SUPERSCRIPT = '\033[73m'
    SUBSCRIPT = '\033[74m'
    NOT_SUPER_SUB = '\033[75m'
    BRIGHT_BLACK = '\033[90m'
    BRIGHT_RED = '\033[91m'
    BRIGHT_GREEN = '\033[92m'
    BRIGHT_YELLOW = '\033[93m'
    BRIGHT_BLUE = '\033[94m'
    BRIGHT_MAGENTA = '\033[95m'
    BRIGHT_CYAN = '\033[96m'
    BRIGHT_WHITE = '\033[97m'
    BG_BRIGHT_BLACK = '\033[100m'
    BG_BRIGHT_RED = '\033[101m'
    BG_BRIGHT_GREEN = '\033[102m'
    BG_BRIGHT_YELLOW = '\033[103m'
    BG_BRIGHT_BLUE = '\033[104m'
    BG_BRIGHT_MAGENTA = '\033[105m'
    BG_BRIGHT_CYAN = '\033[106m'
    BG_BRIGHT_WHITE = '\033[107m'
    UN_BLACK = un_color(0, 0, 0)
    UN_RED = un_color(170, 0, 0)
    UN_GREEN = un_color(0, 170, 0)
    UN_YELLOW = un_color(170, 85, 0)
    UN_BLUE = un_color(0, 0, 170)
    UN_MAGENTA = un_color(170, 0, 170)
    UN_CYAN = un_color(0, 170, 170)
    UN_WHITE = un_color(170, 170, 170)
    UN_BRIGHT_BLACK = un_color(85, 85, 85)
    UN_BRIGHT_RED = un_color(255, 85, 85)
    UN_BRIGHT_GREEN = un_color(85, 255, 85)
    UN_BRIGHT_YELLOW = un_color(255, 255, 85)
    UN_BRIGHT_BLUE = un_color(85, 85, 255)
    UN_BRIGHT_MAGENTA = un_color(255, 85, 255)
    UN_BRIGHT_CYAN = un_color(85, 255, 255)
    UN_BRIGHT_WHITE = un_color(255, 255, 255)

    @staticmethod
    def get(code: Union['AnsiCodes', str]) -> str:
        """ Return the ANSI code from its code name.

        :param code: An instance of AnsiCodes, the ANSI code name or ANSI color code value.
        :return: A string with the ANSI code value.
        """
        if isinstance(code, AnsiCodes):
            return code.value
        if is_color(code):
            return code
        try:
            return AnsiCodes(code).value
        except ValueError:
            return AnsiCodes[code.upper()].value

    def __contains__(self, code: Union['AnsiCodes', str]) -> bool:
        """ Check if a code is a set ANSI code.

        :param code: An instance of AnsiCodes or the ANSI code name.
        :return: True if it is an ANSI code, False otherwise.
        """
        return isinstance(code, AnsiCodes) or bool(AnsiCodes.get(code))


def markup(text: str, *styles: Union[AnsiCodes, str]) -> str:
    """ The same text but with the ANSI codes for the effects.

    :param text: The text to markup.
    :param styles: The styles.
    :return:
    """
    marks = [AnsiCodes.get(style) for style in styles]
    return ''.join(marks) + text + AnsiCodes.NORMAL.value


def is_color(color_code: str) -> bool:
    """ Check if a color code is a valid ANSI color code.
    :param color_code: The color code to evaluate.
    :return: True if the color is valid, False otherwise.
    """
    return bool(re.match(r'^\033\[(([3-4][0-7]|39|49|59)|(38|48|58);(5;[0-9]+|2;[0-9]+;[0-9]+;[0-9]+))m$', color_code))



from json import dumps, loads
from typing import Union

try:
    from requests import post, RequestException
except ModuleNotFoundError as e:
    raise ModuleNotFoundError('ModuleNotFoundError: No module named \'requests\'. '
                              'Please install it with the command:\n\n'
                              'pip install requests~=2.25.1')


def json_post(host: str, msg: Union[list, dict, str]) -> Union[list, dict, str]:
    """ Makes a http json post.
    :param host: The host.
    :param msg: Object to send to the server.
    """
    res = post(host, dumps(msg), headers={'content-type': 'application/json'})
    if not res.ok:
        raise RequestException(f'Error {res.status_code} connecting with "{host}":\n{res.content.decode("utf-8")}')
    return loads(res.content)



import gzip
import pickle
import os
import re
from datetime import datetime
from io import DEFAULT_BUFFER_SIZE
from json import dump, load
from os import makedirs, remove, rmdir, scandir, PathLike
from os.path import exists, dirname, join, basename, isdir
from shutil import copyfile, rmtree
from sys import stdout
from shutil import move
from typing import Union, Optional, TextIO, Any, List, Tuple

from typing.io import IO


def open_file(filename: Union[PathLike, str, bytes],
              mode: str = 'rt',
              buffering: int = DEFAULT_BUFFER_SIZE,
              encoding: Optional[str] = None,
              errors: Optional[str] = None,
              newline: Optional[str] = None,
              close_fd: bool = True,
              opener: Optional = None) -> Union[IO, TextIO]:
    """ Open file and return a stream. Raise OSError upon failure.
    This function is the same as open() but it is able to open a gzip file automatically only taking into account the
    file extension. That means, if the file ends with a .gz extension, then this function will open a gzip file instead
    the normal one.

    :param filename: is either a text or byte string giving the name (and the path if the file isn't in the current
      working directory) of the file to be opened or an integer file descriptor of the file to be wrapped.
      See the open() function documentation for more information.
    :param mode: s an optional string that specifies the mode in which the file is opened. It defaults to 'r' which
      means open for reading in text mode. Other common values are 'w' for writing (truncating the file if it already
      exists), 'x' for creating and writing to a new file, and 'a' for appending (which on some Unix systems, means that
      all writes append to the end of the file regardless of the current seek position).
      See open() function for more information.
    :param buffering: this parameter is ignored for gzip files. It is an optional integer used to set the buffering
      policy. See open() function for more information.
    :param encoding: is the name of the encoding used to decode or encode the file. This should only be used in text
      mode. The default encoding is platform dependent, but any encoding supported by Python can be passed. See the
      codecs module for the list of supported encodings.
    :param errors: errors is an optional string that specifies how encoding errors are to be handled—this argument
      should not be used in binary mode. Pass 'strict' to raise a ValueError exception if there is an encoding error
      (the default of None has the same effect), or pass 'ignore' to ignore errors. (Note that ignoring encoding errors
      can lead to data loss.) See the documentation for codecs.register or run 'help(codecs.Codec)' for a list of the
      permitted encoding error strings.
    :param newline: newline controls how universal newlines works (it only applies to text mode).
      See open() function for more information.
    :param close_fd: If close_fd is False, the underlying file descriptor will be kept open when the file is closed.
      This does not work when a file name is given and must be True in that case.
    :param opener: A custom opener can be used by passing a callable as opener. The underlying file descriptor for the
      file object is then obtained by calling opener with (file, flags). opener must return an open file descriptor
      (passing os.open as opener results in functionality similar to passing None).
      See open() function for more information.
    :return: The opened stream.
    """
    if str(filename).lower().endswith('.gz') or str(filename).lower().endswith('.tgz'):
        return gzip.open(filename, mode, encoding=encoding, errors=errors, newline=newline)
    return open(filename, mode, buffering, encoding, errors, newline, close_fd, opener)


def force_open(filename: Union[PathLike, str, bytes],
               mode: str = 'rt',
               buffering: int = DEFAULT_BUFFER_SIZE,
               encoding: Optional[str] = None,
               errors: Optional[str] = None,
               newline: Optional[str] = None,
               close_fd: bool = True,
               opener: Optional = None) -> Union[IO, TextIO]:
    """ Open file and return a stream. Raise OSError upon failure.
    This function is the same as open_file() but if the file folder does not exist, then create all the necessary
    folders before opening the file.

    :param filename: is either a text or byte string giving the name (and the path if the file isn't in the current
      working directory) of the file to be opened or an integer file descriptor of the file to be wrapped.
      See the open() function documentation for more information.
    :param mode: s an optional string that specifies the mode in which the file is opened. It defaults to 'r' which
      means open for reading in text mode. Other common values are 'w' for writing (truncating the file if it already
      exists), 'x' for creating and writing to a new file, and 'a' for appending (which on some Unix systems, means that
      all writes append to the end of the file regardless of the current seek position).
      See open() function for more information.
    :param buffering: this parameter is ignored for gzip files. It is an optional integer used to set the buffering
      policy. See open() function for more information.
    :param encoding: is the name of the encoding used to decode or encode the file. This should only be used in text
      mode. The default encoding is platform dependent, but any encoding supported by Python can be passed. See the
      codecs module for the list of supported encodings.
    :param errors: errors is an optional string that specifies how encoding errors are to be handled—this argument
      should not be used in binary mode. Pass 'strict' to raise a ValueError exception if there is an encoding error
      (the default of None has the same effect), or pass 'ignore' to ignore errors. (Note that ignoring encoding errors
      can lead to data loss.) See the documentation for codecs.register or run 'help(codecs.Codec)' for a list of the
      permitted encoding error strings.
    :param newline: newline controls how universal newlines works (it only applies to text mode).
      See open() function for more information.
    :param close_fd: If close_fd is False, the underlying file descriptor will be kept open when the file is closed.
      This does not work when a file name is given and must be True in that case.
    :param opener: A custom opener can be used by passing a callable as opener. The underlying file descriptor for the
      file object is then obtained by calling opener with (file, flags). opener must return an open file descriptor
      (passing os.open as opener results in functionality similar to passing None).
      See open() function for more information.
    :return: The opened stream.
    """
    if not exists(dirname(filename)):
        makedirs(dirname(filename))
    return open_file(filename, mode, buffering, encoding, errors, newline, close_fd, opener=opener)


def copy_files(dest: Union[PathLike, str, bytes], *files: Union[PathLike, str, bytes], force: bool = True) -> None:
    """ Copy a list of files into destination folder.
    :param dest: The destination folder.
    :param files: The list of files to copy.
    :param force: Force the creation of the path folders if they do not exist.
    """
    if not exists(dest) and force:
        makedirs(dest)
    for file in files:
        copyfile(file, join(dest, basename(file)))


def save_json(obj: Any, filename: Union[PathLike, str, bytes], force: bool = False) -> None:
    """ Save an object into a json file.
    :param obj: The object to save.
    :param filename: The path to the output file.
    :param force: Force the creation of the path folders if they do not exist.
    """
    with force_open(filename, 'wt') if force else open_file(filename, 'wt') as file:
        dump(obj, file, indent=2)


def load_json(filename: Union[PathLike, str, bytes]) -> Any:
    """ Load a json file and return a object with its data.
    :param filename: The json file.
    :return: An object with the json data.
    """
    with open_file(filename, 'rt') as file:
        return load(file)


def save_pickle(obj: object, filename: Union[PathLike, str, bytes], force: bool = False) -> None:
    """ Save an object into a pickle file.
    :param obj: The object to save.
    :param filename: The path to the output file.
    :param force: Force the creation of the path folders if they do not exist.
    """
    with force_open(filename, 'wb') if force else open_file(filename, 'wb') as file:
        pickle.dump(obj, file)


def load_pickle(filename: Union[PathLike, str, bytes]) -> Any:
    """ Load an object from pickle file.
    :param filename: The pickle file path.
    :return: An object with the pickle file data.
    """
    with open_file(filename, 'rb') as file:
        return pickle.load(file)


def gzip_decompress(input_file: Union[PathLike, str, bytes], output_file: Union[PathLike, str, bytes]) -> None:
    """ Decompress a file using gzip compression.
    :param input_file: The file to compress.
    :param output_file: The compressed file with all the information of the input file.
    """
    if input_file == output_file:
        raise ValueError('The input file and the output file must be different.')
    with open(output_file, 'wb') as writer:
        with gzip.open(input_file, 'rb') as reader:
            for chunk in reader:
                writer.write(chunk)


def gzip_compress(input_file: Union[PathLike, str, bytes], output_file: Union[PathLike, str, bytes]) -> None:
    """ Compress a file using gzip compression.
    :param input_file: The file to compress.
    :param output_file: The compressed file with all the information of the input file.
    """
    if input_file == output_file:
        raise ValueError('The input file and the output file must be different.')
    with gzip.open(output_file, 'wb') as writer:
        with open(input_file, 'rb') as reader:
            for chunk in reader:
                writer.write(chunk)


def remove_files(*files: Union[PathLike, str, bytes], ignore_errors: bool = False, recursive: bool = False) -> None:
    """ Remove several files and empty directories at once.

    :param files: The list of files or empty directories to remove. To remove directories with files or subdirectories,
      please, use shutil.rmtree().
    :param ignore_errors: If True, ignore the error if the file or directory does not exist.
    """
    for file in files:
        if isdir(file):
            if recursive:
                rmtree(file)
            else:
                rmdir(file)
        elif not ignore_errors or exists(file):
            remove(file)


def first_line(filename: Union[PathLike, str, bytes]) -> str:
    """ Read the first line of a file removing the final \n if it exists.

    :param filename: The filename to read.
    :return: A string with the first line.
    """
    with open_file(filename) as file:
        line = file.readline()
        return line[:-1] if line.endswith('\n') else line


def last_line(filename: Union[PathLike, str, bytes]) -> str:
    """ Read the last line of a file removing the final \n if it exists.

    :param filename: The filename to read.
    :return: A string with the last line.
    """
    return read_file(filename, False)[-1]


def head(filename: Union[PathLike, str, bytes], n: int = 10) -> List[str]:
    """ Return a list of with the first n lines of the file.
    :param filename: The file.
    :param n: The number of lines.
    :return: A list of string with the top n lines of the file without the \n.
    """
    with open_file(filename) as file:
        lines = []
        for i, line in enumerate(file.readlines()):
            lines.append(line[:-1] if line.endswith('\n') else line)
            if i >= n - 1:
                return lines
        return lines


def body(filename: Union[PathLike, str, bytes], init: int, n: int = 10) -> List[str]:
    """ Return a part of a text file from a line to another.
    :param filename: The file.
    :param init: The initial line to start reading.
    :param n: The number of lines to read.
    :return: A list of string with the lines between init and init + n without the \n.
    """
    with open_file(filename) as file:
        lines = []
        for i, line in enumerate(file.readlines()):
            if i >= init:
                lines.append(line[:-1] if line.endswith('\n') else line)
            if i >= (init + n) - 1:
                return lines
        return lines


def tail(filename: Union[PathLike, str, bytes], n: int = 10) -> List[str]:
    """ Return a list of with the last n lines of the file.
    :param filename: The file.
    :param n: The number of lines.
    :return: A list of string with the last n lines of the file without the \n.
    """
    return read_file(filename, False)[-n:]


def exist_files(*files: Union[PathLike, str, bytes]) -> bool:
    """ Check if a sequence of files exist.

    :param files: The list of file paths to check.
    :return: True if all the files exist, otherwise False.
    """
    for file in files:
        if not exists(file):
            return False
    return True


def not_exist_files(*files: Union[PathLike, str, bytes]) -> bool:
    """ Check if any of the files exist.

    :param files: The list of file paths to check.
    :return: True if any of the files exist. If exists at least one, then False.
    """
    for file in files:
        if exists(file):
            return False
    return True


def are_dir(*files: Union[PathLike, str, bytes]) -> bool:
    """ Check if a sequence of files are directories.

    :param files: The list of file paths to check.
    :return: True if all of the files are directories, otherwise False.
    """
    for file in files:
        if not isdir(file):
            return False
    return True


def not_are_dir(*files: Union[PathLike, str, bytes]) -> bool:
    """ Check if any of a sequence of files are directories.

    :param files: The list of file paths to check.
    :return: True if any of the files exist, otherwise False.
    """
    for file in files:
        if isdir(file):
            return False
    return True


def count_lines(*filenames: Union[PathLike, str, bytes]) -> int:
    """ Calculate the number of lines in a file.

    :param filenames: The list of filenames to calculate its size.
    :return: The number of lines of all the files.
    """
    count = 0
    for filename in filenames:
        with open_file(filename) as file:
            for _ in file:
                count += 1
    return count


def touch(*files: Union[PathLike, str, bytes]) -> Tuple[Union[PathLike, str, bytes]]:
    """ Create several empty files.

    :param files: The list of file paths to create.
    :return: The list of created files.
    """
    for file in files:
        open(file, 'w').close()
    return files


def cat(filename: Union[PathLike, str, bytes], output: TextIO = stdout) -> None:
    """ Print a file content.

    :param filename: The path to the file. If the file name ends with ".gz", this function decompressed it to print.
    :param output: The stream to print. By default, the standard output.
    """
    with open_file(filename, 'rt') as file:
        for line in file:
            print(line, end='', file=output)


def read_file(filename: Union[PathLike, str, bytes], line_break: bool = True) -> List[str]:
    """ Read a file (compressed with gzip or not) and return in a list its content, each line in a list element.

    :param filename: The path to the file. If the file name ends with ".gz", this function decompressed it first.
    :param line_break: If True, the newline character is conserved, otherwise is removed.
    :return: An array with the contents of the file.
    """
    with open_file(filename, 'rt') as file:
        return [line[:-1] if not line_break and line[-1] == '\n' else line for line in file]


def read_files(*filenames: Union[PathLike, str, bytes], line_break: bool = True) -> List[str]:
    """ Read a file (compressed with gzip or not) and return in a list its content, each line in a list element.

    :param filename: The path to the file. If the file name ends with ".gz", this function decompressed it first.
    :param line_break: If True, the newline character is conserved, otherwise is removed.
    :return: An array with the contents of the file.
    """
    result = []
    for filename in filenames:
        result.extend(read_file(filename, line_break))
    return result


def read_from(filename: Union[PathLike, str, bytes],
              regex: str = '',
              ignore_case: bool = False,
              line_break: bool = True) -> List[str]:
    """ Read from the line that matches with a regular regular.

    :param filename: The path to the file.
    :param regex: The regular expression.
    :param ignore_case: If ignore case or not.
    :param line_break: If True, the newline character is conserved, otherwise is removed.
    :return: A list of strings with each file line.
    """
    pattern = re.compile(regex, flags=re.IGNORECASE if ignore_case else 0)
    from_detected, lines = not regex, []
    with open_file(filename, 'rt') as file:
        for line in file:
            from_detected = from_detected or pattern.match(line)
            if from_detected:
                lines.append(line[:-1] if not line_break and line[-1] == '\n' else line)
    return lines


def read_until(filename: Union[PathLike, str, bytes], regexp: str = '',
               ignore_case: bool = False, line_break: bool = True) -> List[str]:
    """ Read until the line that matches with a regular regular.

    :param filename: The path to the file.
    :param regexp: The regular expression.
    :param ignore_case: If ignore case or not.
    :param line_break: If True, the newline character is conserved, otherwise is removed.
    :return: A list of strings with each file line.
    """
    pattern = re.compile(regexp, flags=re.IGNORECASE if ignore_case else 0)
    until_detected, lines = False, []
    with open_file(filename, 'rt') as file:
        for line in file:
            until_detected = until_detected or regexp and pattern.match(line)
            if not until_detected:
                lines.append(line[:-1] if not line_break and line[-1] == '\n' else line)
    return lines


def mkdirs(*paths: Union[PathLike, str, bytes], mode: int = 0o777,
           dir_fd: int = None) -> Tuple[Union[PathLike, str, bytes]]:
    """ Create one or several directories ignoring the error if the file or folder already exists.
    :param paths: The list of path to the directories.
    :param mode: The mode argument is ignored on Windows. By default, 0o777.
    :param dir_fd: If dir_fd is not None, it should be a file descriptor open to a directory,
        and path should be relative; path will then be relative to that directory.
        dir_fd may not be implemented on your platform.
        If it is unavailable, using it will raise a NotImplementedError.
    :return: The list of created directories.
    """
    for path in paths:
        if not exists(path):
            os.mkdir(path, mode, dir_fd=dir_fd)
    return paths


def move_files(dest: Union[PathLike, str, bytes], *files: Union[PathLike, str, bytes],
               force: bool = False, replace: bool = False) -> None:
    """ Move several files at once.

    :param dest: The destination folder.
    :param files: The files to move.
    :param force: If True, create the folder if it doesn't exist.
    :param replace: if any of the files exist, replace them.
    """
    if force:
        mkdirs(dest)
    for file in files:
        if exists(join(dest, file)) and replace:
            remove(join(dest, file))
        move(file, dest)


def list_dir(folder: Union[PathLike, str, bytes] = '.', filter: str = None, reverse: bool = False) -> List[str]:
    """ List a directory and return a list with all file path of that directory that satisfy the given filter,
        ordered alphabetically.

    :param folder: The folder to list.
    :param filter: The filter to apply.
    :param reverse: If True, the list is inverted sorted. By default, False.
    :return: The list with the path to each directory.
    """
    return sorted([
        join(folder, file.name) for file in scandir(folder) if not filter or re.match(filter, file.name)
    ], reverse=reverse)


def first_file(folder: Union[PathLike, str, bytes] = '.', filter: str = None) -> str:
    """ Obtain the first file name ordered alphabetically from a folder. By default, it uses the current folder.
    :param folder: The folder path with the files.
    :param filter: A regular expression pattern to filter the files. By default, all files are taking into account.
    :return: The first file name.
    """
    files = list_dir(folder, filter)
    return files[0] if files else None


def last_file(folder: Union[PathLike, str, bytes] = '.', filter: str = None) -> str:
    """ Obtain the last file name ordered alphabetically from a folder. By default, it uses the current folder.
    :param folder: The folder path with the files. By default, all files are taking into account.
    :param filter: A regular expression pattern to filter the files.
    :return: The last file name.
    """
    files = list_dir(folder, filter, True)
    return files[0] if files else None


def output_file_path(folder: Union[PathLike, str, bytes] = '.', suffix: str = '', timestamp: bool = True, **kwargs) -> str:
    """ Build a file path in the specified folder based on the current timestamp and several attributes.
    :param folder: The directory where the file will be located.
    :param timestamp: If you want to add the timestamp to the file name.
    :param suffix: If suffix file name.
    :param kwargs: Extra arguments to form the name. If it is a str, float or int value, then the value will be add to
      the name. If it is a bool value and it is True, then the name of the argument is add to the file name.
    :return: The full path to the file to create.
    """
    name = []
    if timestamp:
        name.append(datetime.now().strftime('%Y%m%d-%H%M%S'))
    for param, value in kwargs.items():
        if isinstance(value, bool):
            name += [param] if value else []
        elif value:
            name.append(str(value))

    return join(folder, f'{"-".join(name)}{suffix}')



import re
from logging import getLogger
from os.path import exists

from mysutils.file import read_from
from mysutils.service import endpoint

try:
    from markdown import markdown
except ImportError:
    raise ImportError('To use these functions you need to import the markdown module:\n\n'
                      'pip install markdown==3.3.6\n\n'
                      'Optionally, if you want colourful code, you also need to install Pygments:\n\n'
                      'pip install Pygments>=2.10.0,~=2.11.2')

logger = getLogger(__name__)
try:
    import pygments
except ImportError:
    logger.warning('If you wants to colourful code, please install Pygments:\n\npip install Pygments>=2.10.0,~=2.11.2')

README = 'README.md'


def gen_service_help(title: str, file: str = README, regex: str = '', *endpoints: str) -> str:
    """ Generate a HTML page from a README file or other Markdown file.

    :param title: The page title.
    :param file: The Markdown file.
    :param regex: A regular expression that determines from which line the markdown has to be processed.
    :param endpoints: The different endpoints of the service.
    :return: The Web page.
    """
    help_text = html_header(title)
    help_text += f'<h1>{title}</h1>' \
                 f'<p><b>Note: </b>Apart of this online help, you may use the <a href="docs">API documentation</a> or ' \
                 f'<a href="redoc">API specification</a>.</p>\n'
    if exists(file):
        readme_text = ''.join(read_from(file, regex)[1:])
        help_text += markdown(readme_text, extensions=['fenced_code', 'codehilite'])
        for e in endpoints:
            help_text = re.sub(f'https?://.*{e}', f'{endpoint(e)}', help_text)
    else:
        logger.warning(f'If you want to show an online help, please, add the file {file} to the project root.')

    help_text += '<p>For more information, see the <a href="docs">API documentation</a> or ' \
                 '<a href="redoc">API specification</a>.</p>\n'
    return help_text + html_footer()


def html_header(title: str) -> str:
    """ Generate a Web page header.
    :param title: The title of the page.
    :return: The HTML of the header.
    """
    return '''<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">''' + \
    f'<title>{title}</title>' + \
    '''<style>
        body {
            font-family: system-ui, -apple-system, "Segoe UI", Roboto, "Helvetica Neue", Arial, "Noto Sans", "Liberation Sans", sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol", "Noto Color Emoji";
            color: rgb(33, 37, 41);
            font-size: 16px;
            font-weight: 400;
            line-height: 24px;
            text-align: start;
        }
        h1 {
            box-sizing: border-box;
            font-size: 40px;
            font-weight: 500;
            line-height: 48px;
        }
        h2 {
            font-size: 24px;
            font-weight: 700;
        }
        h3 { font-size: 18.72px; }
        h4 { font-size: 16px; }
        .container {
            box-sizing: border-box;
            margin-left: 34.5px;
            margin-right: 34.5px;
            max-width: 1140px;
            padding-left: 12px;
            padding-right: 12px;
            width: 1140px;
        }
        pre { line-height: 125%; }
        td.linenos .normal { color: inherit; background-color: transparent; padding-left: 5px; padding-right: 5px; }
        span.linenos { color: inherit; background-color: transparent; padding-left: 5px; padding-right: 5px; }
        td.linenos .special { color: #000000; background-color: #ffffc0; padding-left: 5px; padding-right: 5px; }
        span.linenos.special { color: #000000; background-color: #ffffc0; padding-left: 5px; padding-right: 5px; }
        .hll { background-color: #ffffcc }
        { background: #f8f8f8; }
        .c { color: #408080; font-style: italic } /* Comment */
        .err { border: 1px solid #FF0000 } /* Error */
        .k { color: #008000; font-weight: bold } /* Keyword */
        .o { color: #666666 } /* Operator */
        .ch { color: #408080; font-style: italic } /* Comment.Hashbang */
        .cm { color: #408080; font-style: italic } /* Comment.Multiline */
        .cp { color: #BC7A00 } /* Comment.Preproc */
        .cpf { color: #408080; font-style: italic } /* Comment.PreprocFile */
        .c1 { color: #408080; font-style: italic } /* Comment.Single */
        .cs { color: #408080; font-style: italic } /* Comment.Special */
        .gd { color: #A00000 } /* Generic.Deleted */
        .ge { font-style: italic } /* Generic.Emph */
        .gr { color: #FF0000 } /* Generic.Error */
        .gh { color: #000080; font-weight: bold } /* Generic.Heading */
        .gi { color: #00A000 } /* Generic.Inserted */
        .go { color: #888888 } /* Generic.Output */
        .gp { color: #000080; font-weight: bold } /* Generic.Prompt */
        .gs { font-weight: bold } /* Generic.Strong */
        .gu { color: #800080; font-weight: bold } /* Generic.Subheading */
        .gt { color: #0044DD } /* Generic.Traceback */
        .kc { color: #008000; font-weight: bold } /* Keyword.Constant */
        .kd { color: #008000; font-weight: bold } /* Keyword.Declaration */
        .kn { color: #008000; font-weight: bold } /* Keyword.Namespace */
        .kp { color: #008000 } /* Keyword.Pseudo */
        .kr { color: #008000; font-weight: bold } /* Keyword.Reserved */
        .kt { color: #B00040 } /* Keyword.Type */
        .m { color: #666666 } /* Literal.Number */
        .s { color: #BA2121 } /* Literal.String */
        .na { color: #7D9029 } /* Name.Attribute */
        .nb { color: #008000 } /* Name.Builtin */
        .nc { color: #0000FF; font-weight: bold } /* Name.Class */
        .no { color: #880000 } /* Name.Constant */
        .nd { color: #AA22FF } /* Name.Decorator */
        .ni { color: #999999; font-weight: bold } /* Name.Entity */
        .ne { color: #D2413A; font-weight: bold } /* Name.Exception */
        .nf { color: #0000FF } /* Name.Function */
        .nl { color: #A0A000 } /* Name.Label */
        .nn { color: #0000FF; font-weight: bold } /* Name.Namespace */
        .nt { color: #008000; font-weight: bold } /* Name.Tag */
        .nv { color: #19177C } /* Name.Variable */
        .ow { color: #AA22FF; font-weight: bold } /* Operator.Word */
        .w { color: #bbbbbb } /* Text.Whitespace */
        .mb { color: #666666 } /* Literal.Number.Bin */
        .mf { color: #666666 } /* Literal.Number.Float */
        .mh { color: #666666 } /* Literal.Number.Hex */
        .mi { color: #666666 } /* Literal.Number.Integer */
        .mo { color: #666666 } /* Literal.Number.Oct */
        .sa { color: #BA2121 } /* Literal.String.Affix */
        .sb { color: #BA2121 } /* Literal.String.Backtick */
        .sc { color: #BA2121 } /* Literal.String.Char */
        .dl { color: #BA2121 } /* Literal.String.Delimiter */
        .sd { color: #BA2121; font-style: italic } /* Literal.String.Doc */
        .s2 { color: #BA2121 } /* Literal.String.Double */
        .se { color: #BB6622; font-weight: bold } /* Literal.String.Escape */
        .sh { color: #BA2121 } /* Literal.String.Heredoc */
        .si { color: #BB6688; font-weight: bold } /* Literal.String.Interpol */
        .sx { color: #008000 } /* Literal.String.Other */
        .sr { color: #BB6688 } /* Literal.String.Regex */
        .s1 { color: #BA2121 } /* Literal.String.Single */
        .ss { color: #19177C } /* Literal.String.Symbol */
        .bp { color: #008000 } /* Name.Builtin.Pseudo */
        .fm { color: #0000FF } /* Name.Function.Magic */
        .vc { color: #19177C } /* Name.Variable.Class */
        .vg { color: #19177C } /* Name.Variable.Global */
        .vi { color: #19177C } /* Name.Variable.Instance */
        .vm { color: #19177C } /* Name.Variable.Magic */
        .il { color: #666666 } /* Literal.Number.Integer.Long */
    </style>
</head>
<body>
  <div class="container">'''


def html_footer():
    """ Generate a Web page footer.
    :return: The HTML of the Web footer.
    """
    return '</div>\n</body>'



from typing import Callable, Any


def conditional(func: Callable, condition: Any, *args: Any, **kwargs) -> Any:
    """ Call a function with its arguments if the condition is True.

    :param func: The function to execute.
    :param condition: The condition.
    :param args: The function arguments.
    :param kwargs: The function kwargs.
    :return: The same that the function returns.
    """
    if condition:
        return func(*args, **kwargs)


from os import PathLike
from typing import Union

from tqdm.auto import tqdm

try:
    import requests
except ModuleNotFoundError as e:
    raise ModuleNotFoundError('ModuleNotFoundError: No module named \'requests\'. '
                              'Please install it with the command:\n\n'
                              'pip install requests~=2.25.1')


def download(url: str, filename: Union[str, PathLike, bytes], verbose: bool = True) -> None:
    """  Download a file.

    :param url: The URL where the file should be downloaded.
    :param filename: The file path where the file should be stored.
    :param verbose: True, if a progress bar is shown. Otherwise, False.
    """
    request = requests.get(url, stream=True)
    total = int(request.headers['content-length']) if 'content-length' in request.headers else 0
    with request as reader:
        with open(filename, 'wb') as writer:
            with tqdm(desc=f'Downloading file {filename}', total=total, disable=not verbose) as t:
                for chunk in reader.iter_content(chunk_size=8192):
                    t.update(len(chunk))
                    writer.write(chunk)



import gzip
import json
import pickle
import tarfile
from os import makedirs, PathLike
from tarfile import TarInfo
from os.path import basename, isdir, join, exists, splitext, dirname, normpath
from typing import List, Any, Union
from shutil import move

from typing.io import IO


from mysutils.file import copy_files, list_dir
from mysutils.tmp import removable_tmp
# Import tqdm if it is installed, otherwise a dummy tqdm function is used.
try:
    from tqdm.auto import tqdm
except ModuleNotFoundError as e:
    def tqdm(obj: Any) -> Any:
        return obj

# Available tar compress methods
COMPRESS_METHODS = {'gz', 'bz2', 'xz'}


def create_tar(filename: Union[str, PathLike, bytes],
               *files: Union[str, PathLike, bytes],
               verbose: bool = False, compress_method: str = None) -> str:
    """
    Create a tar file with a given list of files. If the filename has any of these extensions 'gz', 'bz2' or 'xz',
      then the tar file will be compressed with the specify method.
    :param filename: The name of the tar file.
    :param files: The list of file paths to include in the tar file.
    :param verbose: if True, show a bar progress.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    compress_method = compress_method if compress_method else detect_compress_method(filename)
    with tarfile.open(filename, f'w:{compress_method}') as tar:
        for file in tqdm(files, desc='Creating tar file', disable=not verbose):
            tar.add(file, basename(normpath(file)))
    return filename


def add_tar_files(filename: Union[str, PathLike, bytes],
                  *files: Union[str, PathLike, bytes],
                  verbose: bool = False, compress_method: str = None) -> str:
    """ Add files to a previously created tar file.

    :param filename: The TAR file. It may be a compressed one or not.
    :param files: A list of file paths to include in the tar file
    :param verbose: If you want to show a progress bar.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    compress_method = compress_method if compress_method else detect_compress_method(filename)
    if compress_method and exists(filename):
        add_compressed_tar_files(filename, *files, verbose=verbose)
    else:
        with tarfile.open(filename, f'a' if exists(filename) else f'w:{compress_method}') as tar:
            for file in tqdm(files, desc='Adding files to tar', disable=not verbose):
                tar.add(file, basename(normpath(file)))
    return filename


def add_compressed_tar_files(filename: Union[str, PathLike, bytes],
                             *files: Union[str, PathLike, bytes],
                             verbose: bool = False, compress_method: str = None) -> str:
    """ Add files to a previously created tar file.

    :param filename: The TAR file. It has to be a compressed one.
    :param files: A list of file paths to include in the tar file
    :param verbose: If you want to show a progress bar.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    compress_method = compress_method if compress_method else detect_compress_method(filename)
    if compress_method and exists(filename):
        with removable_tmp(True) as tmp_dir:
            extract_tar(filename, tmp_dir, False, verbose)
            copy_files(tmp_dir, *files, force=True)
            with removable_tmp(suffix=f'.{compress_method}') as tmp_file:
                create_tar(tmp_file, *list_dir(tmp_dir), verbose=verbose)
                move(tmp_file, filename)
    else:
        with tarfile.open(filename, f'a' if exists(filename) else f'w:{compress_method}') as tar:
            for file in tqdm(files, desc='Adding files to tar', disable=not verbose):
                tar.add(file, basename(file))
    return filename


def detect_compress_method(filename: Union[str, PathLike, bytes]) -> str:
    """ Detecting the compression method based on the extension of the filename.

    :param filename: The file path to the tar file.
    :return: 'gz', 'bz2' or 'xz' if the file is compressed by one of these methods, otherwise an empty string.
    """
    extension = splitext(str(filename))[1]
    if extension.endswith('.tgz'):
        return 'gz'
    return extension[1:].lower() if extension[1:].lower() in COMPRESS_METHODS else ''


def open_tar_file(tar_file: Union[str, PathLike, bytes],
                  filename: Union[str, PathLike, bytes],
                  compress_method: str = None) -> IO:
    """ Open a tar file and return a IO stream to the file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: An IO stream.
    """
    compress_method = compress_method if compress_method else detect_compress_method(tar_file)
    tar = tarfile.open(tar_file, f'r:{compress_method}')
    file = tar.extractfile(filename)
    old_close = file.close

    def close():
        tar.close()
        old_close()

    file.close = close
    return file


def load_tar_json(tar_file: Union[str, PathLike, bytes],
                  filename: Union[str, PathLike, bytes],
                  compress_method: str = None) -> Any:
    """ Load an object from a JSON file stored in a tar file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The loaded object.
    """
    with open_tar_file(tar_file, filename, compress_method) as file:
        if str(filename).lower().endswith('.gz') or str(filename).lower().endswith('.tgz'):
            return json.load(gzip.open(file))
        return json.load(file)


def load_tar_pickle(tar_file: Union[str, PathLike, bytes],
                    filename: Union[str, PathLike, bytes],
                    compress_method: str = None) -> Any:
    """ Load an object from a pickle file stored in a tar file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The loaded object.
    """
    with open_tar_file(tar_file, filename, compress_method) as file:
        if str(filename).lower().endswith('.gz') or str(filename).lower().endswith('.tgz'):
            return pickle.load(gzip.open(file))
        return pickle.load(file)


def list_tar(tar_file: Union[str, PathLike, bytes], compress_method: str = None) -> List[TarInfo]:
    """ Obtain a list with the information of each file or directory of a tar file.

    :param tar_file: The path to the tar file.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: A list of TarInfo instances.
    """
    compress_method = compress_method if compress_method else detect_compress_method(tar_file)
    with tarfile.open(tar_file, f'r:{compress_method}') as tar:
        return tar.getmembers()


def extract_tar_file(tar_file: Union[str, PathLike, bytes],
                     dest: Union[str, PathLike, bytes],
                     filename: Union[str, PathLike, bytes],
                     compress_method: str = None) -> str:
    """ Extract a file inside of a tar archive.

    :param tar_file: The path to the tar file.
    :param dest: The file or folder where the tar archive should be extracted.
    :param filename: The relative path inside of the tar file to extract.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    compress_method = compress_method if compress_method else detect_compress_method(tar_file)
    dest = join(dest, filename) if isdir(dest) else dest
    with tarfile.open(tar_file, f'r:{compress_method}') as tar:
        with tar.extractfile(filename) as reader:
            if dirname(dest) and not exists(dirname(dest)):
                makedirs(dirname(dest))
            with open(dest, 'wb') as writer:
                for chunk in reader:
                    writer.write(chunk)
    return tar_file


def extract_tar_files(tar_file: Union[str, PathLike, bytes],
                      dest: Union[str, PathLike, bytes],
                      *files: Union[str, PathLike, bytes],
                      force: bool = False, verbose: bool = False, compress_method: str = None) -> str:
    """ Extract a file inside of a tar archive.

    :param tar_file: The path to the tar file.
    :param dest: The folder where the tar archive should be extracted.
    :param files: The relative path inside of the tar file to extract.
    :param force: If True, create the destination directory if it doesn't exist.
      If files is not given, then extract all the files.
    :param verbose: If verbose, show the progress bar.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    if not exists(dest) and force:
        makedirs(dest)
    if not isdir(dest):
        raise ValueError(f'The destination is not an existing folder.')
    files = files if files else [file.path for file in list_tar(tar_file)]
    for file in tqdm(files, desc='Extracting files', disable=not verbose):
        extract_tar_file(tar_file, dest, file, compress_method)
    return tar_file


def extract_tar(tar_file: Union[str, PathLike, bytes], dest: Union[str, PathLike, bytes],
                force: bool = False, verbose: bool = False, compress_method: str = None) -> str:
    """ Extract all the files inside of a tar file in the specified directory.

    :param tar_file: The tar file to extract.
    :param dest: The destination directory.
    :param force: If True, create the destination directory if it doesn't exist.
    :param verbose: If verbose, show the progress bar.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: The TAR filename.
    """
    if not exists(dest) and force:
        makedirs(dest)
    if not exists(dest):
        raise FileNotFoundError(f'The folder "{dest}" does not exists. Create it or put the parameter force to True.')

    files = list_tar(tar_file, compress_method)
    for file in tqdm(files, desc='Extracting files', disable=not verbose):
        if file.isdir():
            makedirs(join(dest, file.path))
        else:
            extract_tar_file(tar_file, dest, file.path, compress_method)
    return tar_file


def exist_tar_files(tar_file: Union[str, PathLike, bytes],
                    *files: Union[str, PathLike, bytes],
                    compress_method: str = None) -> bool:
    """ Check if a sequence of files exist.


    :param tar_file: The TAR file.
    :param files: The list of file paths to check.
    :param compress_method: Force the compression or decompression method to use.
       By default, select from the file extension.
    :return: True if all the files exist, otherwise False.
    """
    compress_method = compress_method if compress_method else detect_compress_method(tar_file)
    with tarfile.open(tar_file, f'r:{compress_method}') as tar:
        filenames = tar.getnames()
    for file in files:
        if file not in filenames:
            return False
    return True



import gzip
from collections import OrderedDict, Hashable
from os import PathLike
from typing import Union, Dict, Any

from mysutils.tar import open_tar_file

try:
    from yaml import add_representer, dump, load, SafeLoader
except ModuleNotFoundError as e:
    raise ModuleNotFoundError('ModuleNotFoundError: No module named \'yaml\'. Please install it with the command:\n\n'
                              'pip install PyYAML~=5.4.1')

from mysutils.file import open_file, force_open


def save_yaml(data: Union[Dict[Hashable, Any], list, None],
              filename: Union[str, PathLike, bytes],
              force: bool = False) -> None:
    """ Save an object as a YAML file preserving the dictionary order.
    :param filename: The path to the output file.
    :param data: The data to save.
    :param force: Force the creation of the path folders if they do not exist.
    """
    add_representer(OrderedDict, representer=lambda self, d: self.represent_mapping('tag:yaml.org,2002:map', d.items()))
    with force_open(filename, 'wt') if force else open_file(filename, 'wt') as file:
        dump(data, file, default_flow_style=False)


def load_yaml(filename: Union[str, PathLike, bytes]) -> Union[Dict[Hashable, Any], list, None]:
    """ Load an object from a YAML file.
    :param filename: The path to the YAML file.
    :return: The loaded object.
    """
    with open_file(filename, 'rt') as file:
        return load(file, SafeLoader)


def load_tar_yaml(tar_file: Union[str, PathLike, bytes], filename: Union[str, PathLike, bytes]) -> Any:
    """ Load an object from a YAML file stored in a tar file.

    :param tar_file: The path to the tar file-.
    :param filename: The path inside of the tar to the file to extract.
    :return: The loaded object.
    """
    with open_tar_file(tar_file, filename) as file:
        if str(filename).lower().endswith('.gz') or str(filename).lower().endswith('.tgz'):
            return load(gzip.open(file))
        return load(file)



from typing import Union, List, Any, Callable, Dict, Iterable


def head(obj: Union[dict, set], top: int = 10) -> Union[dict, set]:
    """ Obtain the first top elements of a dictionary o a set.

    :param obj: A dictionary or a set.
    :param top: The number of elements to return, by default 10.
    :return: The first top elements of the given dictionary or set.
    """
    if isinstance(obj, dict):
        return dh(obj, top)
    if isinstance(obj, set):
        return sh(obj, top)
    raise ValueError('The object must be a dictionary or a set.')


def dh(d: dict, top: int = 10) -> dict:
    """ Obtain the first top elements of a dictionary.

    :param d: The dictionary.
    :param top: The number of dictionary items to return
    :return: A dictionary with the top first items of the given dictionary.
    """
    return {key: value for i, (key, value) in enumerate(d.items()) if i < top}


def sh(d: set, top: int = 10) -> set:
    """ Obtain the first top elements of a set.

    :param d: The set.
    :param top: The number of set items to return
    :return: A set with the top first items of the given set.
    """
    return {value for i, value in enumerate(sorted(d)) if i < top}


def list_union(*lists: list) -> list:
    """ Create a new list with the union of the lists without repeated elements.

    :param lists: The list to union.
    :return: The union of all lists respecting the element order.
    """
    result = []
    for l in lists:
        for e in l:
            if e not in result:
                result.append(e)
    return result


def del_keys(d: dict, *keys: Any, ignore_errors: bool = True) -> dict:
    """ Remove the dictionary items from their keys and return the modified dictionary.

    :param d: The dictionary.
    :param keys: A key or a list of keys.
    :param ignore_errors: If True, ignore if the key does not exist.
    :return: The modified dictionary.
    """
    for key in keys:
        if not (key not in d and ignore_errors):
            del d[key]

    return d


def add_keys(d: dict, modify: bool = True, **kwargs) -> dict:
    """ Add or change several items into a dictionary and return the modified dictionary.

    :param d: The dictionary.
    :param modify: If True, if a key already exists, then update it, otherwise raise an error.
    :param kwargs: The dictionary key/value pairs.
    :return: The modified dictionary.
    """
    for key, value in kwargs.items():
        if not modify and key in d:
            raise KeyError("Duplicate key")
        d[key] = value

    return d


def mod_key(d: dict, old_key: Any, new_key: Any) -> dict:
    """ Modify the name of a dictionary key without change anything else.

    :param d: The dictionary to modify.
    :param old_key: The old key to replace for.
    :param new_key: The new key to replace with.
    :return: The modified dictionary.
    """
    value = d[old_key]
    del_keys(d, old_key)[new_key] = value
    return d


def mod_keys(d: dict, **kwargs) -> dict:
    """ Modify the names of a dictionary keys without change anything else with just an instruction.

    :param d: The dictionary to modify.
    :param kwargs: The old keys assigned to the name of the new ones.
    :return: The modified dictionary.
    """
    for old_key, new_key in kwargs.items():
        value = d[old_key]
        del_keys(d, old_key)[new_key] = value
    return d


def mod_value(d: dict, key: Any, value: Any) -> dict:
    """ Modify the value of a dictionary item without change anything else and return the modified dictionary.

    :param d: The dictionary to modify.
    :param key: The item key to modify.
    :param value: The value to replace with.
    :return: The modified dictionary.
    """
    d[key] = value
    return d


def mod_values(d: dict, **kwargs) -> dict:
    """ Modify several value of a dictionary items with just an instruction.

    :param d: The dictionary to modify.
    :param kwargs: The keys and their values to modify.
    :return: The modified dictionary.
    """
    for key, value in kwargs.items():
        d[key] = value
    return d


def filter_lst(lst: list, n: int = 0, init: int = 0, filter_func: Callable = None) -> list:
    """ Filter a list.

    :param lst: The list to filter.
    :param n: The maximum number of results. If it is not given, then return all the results.
    :param init: The initial result. Before this position the items are filtered.
    :param filter_func: A filter function to filter.
    :return: The filtered list.
    """
    n = n if n else len(lst)
    lst = lst[init:init + n]
    return [e for e in filter(filter_func, lst)] if filter_func else lst


def first_key(dicts: Iterable[Dict[Any, Any]], key: Any) -> Any:
    """ From a iterable of dictionaries, return the key value of the first dictionary that contains that key.

    :param dicts: The iterable of dictionaries.
    :param key: The key to search.
    :return: The value of the first dictionary that contains that key.
    :raises KeyError: If any of the dictionaries contains that key.
    """
    for d in dicts:
        if key in d:
            return d[key]
    raise KeyError(f'The key "{key}" does not exist in any of the dictionaries.')


def merge_dicts(dicts: Iterable[Dict[Any, Any]]) -> Dict[Any, List[Any]]:
    """ Convert a list of dictionaries with the same keys in a dictionary which each key contain the list of values.

    :param dicts: The list of dictionaries to merge.
    :return: The dictionary with merged values.
    """
    values = {}
    for d in dicts:
        for key, value in d.items():
            values[key] = values[key] if key in values else []
            values[key].append(value)
    return values


def merge_tuples(tuples: Iterable[tuple]) -> tuple:
    """ Convert a list of dictionaries with the same keys in a dictionary which each key contain the list of values.
       All the tuples has to have the same length.

    :param tuples: The list of tuples to merge.
    :return: The tuple with merged values, in which each element contain a list with the values.
    """
    result = tuple([] for _ in list(tuples)[0]) if tuples else ()
    for t in tuples:
        for i, e in enumerate(t):
            result[i].append(e)
    return result



from os import PathLike
from tempfile import mkdtemp, mktemp
from typing import Tuple, Union, Iterator

from mysutils.file import remove_files


class Removable(object):
    """ Class to include enter and exit methods for removable files. """
    @property
    def files(self) -> Tuple[Union[str, PathLike, bytes]]:
        """
        :return: The list of files to remove.
        """
        return self.__files

    @property
    def ignore_errors(self) -> bool:
        """
        :return: Ignore the error if the file does not exist.
        """
        return self.__ignore_errors

    @property
    def recursive(self) -> bool:
        """
        :return: If True, remove the folders recursively, otherwise False.
        """
        return self.__recursive

    def __init__(self, *files: Union[PathLike, str, bytes], recursive: bool = False,
                 ignore_errors: bool = True) -> None:
        """ Constructor.

        :param files: The files to remove.
        :param ignore_errors: If True, if a file does not exist, then ignore the error.
        :param recursive: If True, remove the folders recursively.
        """
        self.__files = files
        self.__ignore_errors = ignore_errors
        self.__recursive = recursive

    def __enter__(self) -> object:
        """
        :return: return self.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Remove all the removable files.
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        """
        remove_files(*self.__files, ignore_errors=self.__ignore_errors, recursive=self.__recursive)

    def __repr__(self) -> str:
        """
        :return: A string representation of this object.
        """
        return str(self.__files)

    def __getitem__(self, item: int) -> Union[PathLike, str, bytes]:
        """ Obtain an path of the removable files.
        :param item: The index of the file to retrieve.
        :return: The path to the removable file of the index item.
        """
        return self.__files[item]

    def __iter__(self) -> Iterator:
        """
        :return: The iterator of the removable files.
        """
        return iter(self.__files)


def removable_files(*files: Union[PathLike, str, bytes], recursive: bool = False, ignore_errors: bool = True) -> object:
    """ This function is used with "with" python command. As following:

    .. code-block:: python

        from mysutils.file import removable_files, exist_files
        # These files will be removed when the with ends
        with removable_files('test2.json', 'data/test1.json', 'data/'):
            exist_files('test2.json', 'data/test1.json', 'data/')  # Returns True
        exist_files('test2.json', 'data/test1.json', 'data/')  # Returns False

    :param files: The files to remove al the end.
    :param recursive: If True, remove the folders recursively.
    :param ignore_errors: If True, ignore the errors, for example the file not found error.
    :return: A object with enter and exit methods.
    """
    return Removable(*files, recursive=recursive, ignore_errors=ignore_errors)


class RemovableTemp(Removable, PathLike):
    """ Create a removable temporal file. """
    def __init__(self, folder: bool = False, suffix: str = '', prefix: str = ''):
        """

        :param folder:
        :param suffix:
        :param prefix:
        """
        self.__tmp = mkdtemp(suffix=suffix, prefix=prefix) if folder else mktemp(suffix=suffix, prefix=prefix)
        Removable.__init__(self, self.__tmp, recursive=folder, ignore_errors=True)

    def __fspath__(self) -> str:
        """
        :return: The path to the temporal path.
        """
        return self.__tmp

    def __enter__(self) -> object:
        """
        :return: return self.
        """
        return self[0]


def removable_tmp(folder: bool = False, suffix: str = '', prefix: str = '') -> PathLike:
    """ This function is used with "with" python command. As following:

    .. code-block:: python

        from mysutils.file import removable_files, exist_files
        # These files will be removed when the with ends
        with removable_tmp() as tmp:
            exist_files(tmp.files[0])  # Returns True
        exist_files('test2.json', 'data/test1.json', 'data/')  # Returns False

    :return: A object with enter and exit methods.
    """
    return RemovableTemp(folder, suffix, prefix)
